import sys
import os
import logging as lg
import datetime

lg.basicConfig()
logging = lg.getLogger()
logging.setLevel(lg.INFO)
import sqlalchemy


class DB(object):
    """Takes a query string and runs it to see if it returns any rows

            Args:
              table_name (str): String

            Returns:
              boolean: True/False
            """

    def __init__(self):

        self.cursor = None



    def query(self, sql):
        """This will execute a query and fetches all the results and closes the curor.
            this is not meant to be a data store but a utility to pull bits of data from the database

        Args:(self, sql):
          sql (str): Sql compliant sytanx for the specific database type

        Returns:
          None: None
        """
        self.create_cur()
        """ runs query or procedure that returns record set
        """
        logging.debug('Running Query: {}\n\t{}'.format(datetime.datetime.now().time(), sql))
        if sql.lower().startswith('select') or sql.lower().startswith('call'):
            self.cursor.execute(sql)
            rows = self.cursor.fetchall()
            meta = self.cursor.description
            self.cursor.close()
            self.cursor = None

        else:
            raise Exception('Only Selects allowed')
        logging.debug('Query Completed: {}'.format(datetime.datetime.now().time()))
        return rows, meta

    def get_table_columns(self,table_name):
        """This method will select 1 record from the table and return the column names

                Args:(table_name (fully qualified):

                Returns:
                  List: List of Strings
                """
        sql="""select * from {} limit 1""".format(table_name)
        rows,meta=self.query(sql)

        return [str(r[0]) for r in meta]

    def create_cur(self):
        """This method will get called to create a cusor if one does not exist for the instance

        Args:(self):

        Returns:
          None: None
        """
        if self.cursor is None:
            self.cursor = self.conn.cursor()

    def execute(self, sql, commit=False, catch_exception=True):
        """Take a sql string specific to DB instance type and executes

        Args:
          sql (str): String
          commit(Boolean): True/False
          catch_exception(Boolean): True/False

        Returns:
          None: None
        """
        self.create_cur()
        logging.debug(
            "Debug DB Execute: {}: \n\t{} ".format(self.str, sql))
        rowcount = 0
        this_sql = str(sql).strip()

        if catch_exception:
            try:
                self.cursor.execute(this_sql)
            except Exception as e:
                # print("Error Execute SQL:{}".format(e))
                logging.warning("SQL error Occurred But Continuing:\n{}".format(e))
        else:
            self.cursor.execute(this_sql)
        rowcount = self.cursor.rowcount
        if commit:
            self.commit()

        logging.debug("DB SQL Execute Completed: {}".format(self.str))

        return rowcount


    def bulk_load(self):
        pass

    def commit(self):
        """Commits anything from the current open connection and closes the cursor

        Args:(self):

        Returns:
          None: None
        """
        """ Default to commit after every transaction
                Will check instance variable to decide if a commit is needed
        """
        try:
            self.cursor.execute("COMMIT")
            self.cursor.close()
            self.cursor = None
        except AttributeError:
            logging.error("No Open Cursor to do Commit")
        except Exception as e:
            logging.error(e)

    def rollback(self):
        """Rolls back anything the instance has run and closes the cursor

        Args:(self):
          table_name (str): String

        Returns:
          None: None
        """
        self.cursor.execute("ROLLBACK")
        self.cursor.close()
        self.cursor = None

    def query_to_parquet(self, file_path, sql):
        import pyarrow.parquet as pq
        import pyarrow as pa
        import pandas
        df = pandas.read_sql(sql, self.connect_SqlAlchemy())
        print(file_path,"--------------")
        table = pa.Table.from_pandas(df)
        pq.write_table(table, os.path.abspath(file_path))

    def query_to_hdf5(self, file_path, sql, key='table'):
        import pandas

        df = pandas.read_sql(sql, self.connect_SqlAlchemy())
        df.to_hdf(path_or_buf=os.path.abspath(file_path), key=key,mode='w')

    def query_to_csv_pandas(self, file_path, sql, include_header=True):
        import pandas
        df = pandas.read_sql(sql, self.connect_SqlAlchemy())
        df.to_csv(path_or_buf=os.path.abspath(file_path), header=include_header)
    def query_to_csv(self, file_path, sql, header=None):
        import pandas
        rows,meta=self.query(sql)
        df = pandas.DataFrame(data=rows,columns=header)
        df.to_csv(path_or_buf=os.path.abspath(file_path), header=header,index=False)


class ConnRDBMS(object):
    def __init__(self, autocommit=None, pwd=None, userid=None, host=None, dbname=None, schema=None):
        self.str = 'DB: {}:{}:{}:{}:autocommit={}'.format(self.host, self.port, self.dbname, self.userid,
                                                          self.autocommit)
        try:
            self.cursor = None
            self.autocommit = autocommit or True
            print('INIT DB: {}:{}:{}:{}:autocommit={}'.format(self.host, self.port, self.dbname, self.userid,
                                                              self.autocommit))
            logging.debug('Connect: {}:{}:{}'.format(self.host, self.dbname, self.userid))
        except Exception as e:
            logging.debug("Can not Use this Class directly: You must instantiate a child")
            sys.exit(1)


    def __repr__(self):
        return self.str

    def __str__(self):

        return self.str

    def __del__(self):
        logging.debug("Destroying: {}".format(self.str))




    def authenticate(self):
        pass

    def close(self):
        # self.cursor.close()
        self.conn.close()

    def connect_SqlAlchemy(self):
        if self.sql_alchemy_uri is None:
            logging.error("sqlAlchemy not supported for this DB")
            sys.exit(1)
        else:
            try:
                return sqlalchemy.create_engine(self.sql_alchemy_uri.format(
                    userid=self.userid,
                    pwd=self.pwd,
                    host=self.host,
                    port=self.port,
                    db=self.dbname
                ))
            except Exception as e:
                logging.error("Could not Connect to sqlAlchemy, Check Uri Syntax: {}".format(e))
                sys.exit(1)

    def schema_exists(self, schema_name):
        """Takes a query string and runs it to see if it returns any rows

        Args:
          table_name (str): String

        Returns:
          boolean: True/False
        """
        self.create_cur()
        v_found = False

        v_found = self.has_record(
            """select 1 from information_schema.schemata where schema_name='{0}' limit 1""".format(schema_name))

        return v_found

    def table_exists(self, fqn_table_name):
        """Takes a query string and runs it to see if it returns any rows

        Args:
          table_name (str): String

        Returns:
          boolean: True/False
        """
        self.create_cur()

        v_table_exists = False
        v_schema = fqn_table_name.split('.')[0]
        v_table_name = fqn_table_name.split('.')[1]
        v_table_exists = self.has_record(
            """select 1 from information_schema.tables where table_schema='{0}' and table_name='{1}'""".format(v_schema,
                                                                                                               v_table_name))

        return v_table_exists

    def get_a_row(self, sql):
        """Returns 1 row as tuple:
            use var,=row for 1 element tuple
        Args:(self, sql):
          sql (str): String

        Returns:
          row: tuple
        """
        self.create_cur()
        self.cursor.execute(sql)
        row = self.cursor.fetchone()
        self.cursor.close()
        self.cursor = None

        return row

    def has_record(self, sql):
        """Takes a query string and runs it to see if it returns any rows

        Args:
          sql (str): String

        Returns:
          boolean: True/False
        """
        self.create_cur()
        rs = None
        record_len = 0
        try:
            row = self.get_a_row(sql)
        except Exception as e:
            logging.error("error in dbconn.has_record: {}".format(sql))
        record_len = len(row)
        # we need to close this or it will lock the connection

        if record_len > 0:
            return True
        return False

    def create_table_from_dataframe(self, dataframe, table_name_fqn, default_owner=None):
        """Describe Method:

        Args:(self, dataframe, table_name_fqn):
          table_name (str): String

        Returns:
          None: None
        """
        if '.' in table_name_fqn:
            if not self.table_exists(table_name_fqn):
                self.create_cur()
                df = dataframe.head()
                schema = table_name_fqn.split('.')[0]
                table_name = table_name_fqn.split('.')[1]

                engine = self.connect_SqlAlchemy()

                df.to_sql(table_name, engine, schema=schema, if_exists='append', index=False, chunksize=1000)

                self.execute('truncate table {}'.format(table_name_fqn))
                return True
            else:
                logging.info("Table exists: {}".format(table_name_fqn))
                return False
        else:
            logging.error('Please provide fully qualified table name')
            return False

    def table_exists(self, table_name_fqn):
        """Describe Method:

        Args:( table_name_fqn):
          table_name (str): String

        Returns:
          Boolean: True/False
        """
        if self._dbtype == 'POSTGRES':
            table_exists = False
            sql = """select count(*) from information_schema.tables
            WHERE table_type='BASE TABLE'
            and table_name='{table_name}'  
            """
            if '.' in table_name_fqn:
                table_name = table_name_fqn.split('.')[1]
                schema = """ and table_schema='{} limit 1'""".format(table_name_fqn.split('.')[0])

                sql = sql.format(table_name=table_name, schema=schema)
            else:
                return False

            x, = self.get_a_row(sql)
            if x > 0:
                table_exists = True
        else:
            print('Not Supported')
            raise
        return table_exists


class ConnREST(object):
    def __init__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __repr__(self):
        pass

    def __str__(self):
        pass

    def __del__(self):
        pass

    def authenticate(self):
        pass

    def close(self):
        pass
